"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_cv.src.losses.centernet_box_loss import CenterNetBoxLoss
from keras_cv.src.losses.ciou_loss import CIoULoss
from keras_cv.src.losses.focal import FocalLoss
from keras_cv.src.losses.giou_loss import GIoULoss
from keras_cv.src.losses.iou_loss import IoULoss
from keras_cv.src.losses.penalty_reduced_focal_loss import (
    BinaryPenaltyReducedFocalCrossEntropy,
)
from keras_cv.src.losses.simclr_loss import SimCLRLoss
from keras_cv.src.losses.smooth_l1 import SmoothL1Loss
